import base64
import hashlib

from politylink.graphql.schema import *

"""
Minimal content based ID generator

ID = ${class}:${base}
$class = class name
$base = URL-safe base64 string of MD5 hash (length = 22)
"""


def idgen(obj):
    class_ = obj.__class__.__name__
    if isinstance(obj, str):
        base = _basegen_str(obj)
    elif isinstance(obj, Bill):
        base = _basegen_bill(obj)
    elif isinstance(obj, Law):
        base = _basegen_law(obj)
    elif isinstance(obj, Url) or isinstance(obj, News):
        base = _basegen_url(obj.url)
    elif isinstance(obj, Timeline):
        base = _basegen_timeline(obj)
    elif isinstance(obj, Minutes):
        base = _basegen_minutes(obj)
    elif isinstance(obj, Speech):
        base = _basegen_speech(obj)
    elif isinstance(obj, Diet):
        base = _basegen_diet(obj)
    elif isinstance(obj, Activity):
        base = _basegen_activity(obj)
    elif hasattr(obj, 'name'):
        base = _basegen_str(getattr(obj, 'name'))
    else:
        raise ValueError(f'can not calculate ID for {class_}')
    return f'{class_}:{base}'


def _basegen_str(s: str):
    if not s:
        raise ValueError(f'non-empty string is required to calculate ID')
    h = hashlib.md5()
    h.update(s.encode('UTF-8'))
    b = base64.b64encode(h.digest(), altchars=b'-_')
    return b.decode('UTF-8')[:-2]  # last 2 bit is always padding (=)


def _basegen_bill(bill: Bill):
    return _basegen_str(bill.bill_number)


def _basegen_law(law: Law):
    return _basegen_str(law.law_number)


def _basegen_url(url: str):
    protocol, body = url.split('://')
    return _basegen_str(body)


def _basegen_timeline(timeline: Timeline):
    dt = timeline.date
    return f'{dt.year:04}{dt.month:02}{dt.day:02}'


def _basegen_minutes(minutes: Minutes):
    dt = minutes.start_date_time
    dt_str = f'{dt.year:04}{dt.month:02}{dt.day:02}'
    return _basegen_str(minutes.name + dt_str)


def _basegen_speech(speech: Speech):
    return _basegen_str(f'{speech.minutes_id}{speech.order_in_minutes}')


def _basegen_diet(diet: Diet):
    return diet.number


def _basegen_activity(activity: Activity):
    id_str = ''
    for id_field in ['member_id', 'bill_id', 'minutes_id']:  # order matters
        if hasattr(activity, id_field):
            id_str += getattr(activity, id_field)
    dt = activity.datetime
    dt_str = f'{dt.year:04}{dt.month:02}{dt.day:02}'
    return _basegen_str(f'{id_str}{dt_str}')



from typing import List


class AbstractFinder:
    def __init__(self, search_fields):
        self.search_fields = search_fields

    def find(self, text, exact_match=False, *args, **kwargs) -> list:
        NotImplemented

    def find_one(self, text, exact_match=False, *args, **kwargs):
        objects = self.find(text, exact_match, *args, **kwargs)
        if len(objects) == 1:
            return objects[0]
        else:
            raise ValueError(f'{self.__class__.__name__} found {len(objects)} results that match with {text}:{objects}')

    def is_text_match(self, obj, text, exact_match=False) -> bool:
        for field in self.search_fields:
            if hasattr(obj, field):
                for field_text in extract_texts(getattr(obj, field)):
                    if exact_match:
                        if text == field_text:
                            return True
                    else:
                        if text in field_text or field_text in text:
                            return True
        return False


def extract_texts(value) -> List[str]:
    """
    recursively extract texts
    """

    texts = []
    if isinstance(value, str) and value:
        texts.append(value)
    elif isinstance(value, list):
        for e in value:
            texts += extract_texts(e)
    return texts



from os import path

import nltk
import pke
from politylink.graphql.client import GraphQLClient
from spacy.lang.ja import stop_words


class KeyPhraseExtractor:

    def __init__(self):
        # load stop words
        stopwords = set()

        p = path.join(path.dirname(__file__), 'stopwords_minutes.txt')
        stopwords.update(self.load_stopwords(p))

        p = path.join(path.dirname(__file__), 'stopwords_slothlib.txt')
        stopwords.update(self.load_stopwords(p))

        stopwords.update(self.get_member_names())

        stopwords.update(stop_words.STOP_WORDS)
        stopwords = list(stopwords)

        # set stop words
        pke.base.lang_stopwords['ja_ginza'] = 'japanese'
        nltk.corpus.stopwords.words_org = nltk.corpus.stopwords.words
        nltk.corpus.stopwords.words = lambda lang: stopwords if lang == 'japanese' else nltk.corpus.stopwords.words_org(
            lang)

    def load_stopwords(self, file_path):
        with open(file_path, mode='r') as f:
            stopwords = f.read().splitlines()
        return stopwords

    def get_member_names(self):
        client = GraphQLClient()
        fields = ['first_name', 'first_name_hira', 'last_name', 'last_name_hira']
        members = client.get_all_members(fields=fields)
        member_names = [m[f] for f in fields for m in members]
        return member_names

    def extract(self, text, n=3):
        """
        Get key phrases from text
        Up to n key phrases will be returned
        :param text: text
        :param n: maximum number of returned key phrases
        :return: list of key phrases
        """

        extractor = pke.unsupervised.MultipartiteRank()
        extractor.load_document(input=text, language='ja_ginza', normalization=None)
        extractor.candidate_selection(pos={'NOUN', 'PROPN', 'ADJ', 'NUM'})
        extractor.candidate_weighting(threshold=0.74, method='average', alpha=1.1)
        key_phrases = extractor.get_n_best(n)
        key_phrases = [k[0] for k in key_phrases]

        return key_phrases



from functools import partial

from sgqlc.endpoint.http import HTTPEndpoint
from sgqlc.operation import Operation

from politylink.graphql import POLITYLINK_AUTH, POLITYLINK_URL
from politylink.graphql.schema import *
from politylink.graphql.schema import _NewsFilter, _Neo4jDateTimeInput, _ActivityFilter


class GraphQLException(Exception):
    pass


class GraphQLClient:
    """
    GraphQLClient for PolityLink endpoint
    """

    def __init__(self, url=POLITYLINK_URL, auth=POLITYLINK_AUTH, batch_size=100):
        self.endpoint = HTTPEndpoint(url, {'Authorization': auth}, timeout=30)
        self.batch_size = batch_size

    def exec(self, op):
        """
        General method to execute GraphQL operation (query or mutation)
        :param op: GraphQL operation string
        :return: json data
        """

        res = self.endpoint(op)
        self.validate_response_or_raise(res)
        return res['data']

    def get(self, id_, fields=None, depth=2):
        """
        General method to get single GraphQL object by id
        """

        op = self.build_get_operation(id_, fields)
        res = self.endpoint(op.__to_graphql__(auto_select_depth=depth))
        self.validate_response_or_raise(res)
        method_name = id_.split(':')[0].lower()
        data = getattr(op + res, method_name)
        if len(data) == 0:
            raise GraphQLException(f'{id_} does not exist')
        elif len(data) == 1:
            return data[0]
        else:
            raise GraphQLException(f'multiple {id_} exist')

    def merge(self, obj):
        """
        General method to merge GraphQL object
        :param obj: GraphQL object
        :return: json data
        """

        op = self.build_merge_operation(obj)
        return self.exec(op)

    def delete(self, id_):
        """
        General method to delete single GraphQL object by id
        """

        op = self.build_delete_operation(id_)
        return self.exec(op)

    def link(self, from_id, to_id):
        """
        General method to link GraphQL objects by id
        :param from_id: politylink id
        :param to_id: politylink id
        :return: json data
        """

        op = self.build_link_operation(from_id, to_id)
        return self.exec(op)

    def unlink(self, from_id, to_id):
        """
        General method to unlink GraphQL objects by id
        :param from_id: politylink id
        :param to_id: politylink id
        :return: json data
        """

        op = self.build_link_operation(from_id, to_id, remove=True)
        return self.exec(op)

    def bulk_merge(self, objects):
        """
        General method to bulk merge GraphQL objects
        requests will be send in batch to avoid Payload Too Large exception
        :param objects: list of GraphQL objects
        :return: json data of the last request
        """

        op_builder_list = map(lambda x: partial(self.build_merge_operation, obj=x), objects)
        return self.bulk_mutation(op_builder_list)

    def bulk_delete(self, ids):
        """
        General method to bulk delete GraphQL objects by id
        requests will be send in batch to avoid Payload Too Large exception
        :param ids: list of politylink ids
        :return: json data of the last request
        """

        op_builder_list = map(lambda x: partial(self.build_delete_operation, id_=x), ids)
        return self.bulk_mutation(op_builder_list)

    def bulk_link(self, from_ids, to_ids):
        """
        General method to bulk link GraphQL objects by id
        requests will be send in batch to avoid Payload Too Large exception
        :param from_ids: list of politylink ids
        :param to_ids: list of politylink ids
        :return: json data of the last request
        """

        op_builder_list = map(lambda x, y: partial(self.build_link_operation, from_id=x, to_id=y), from_ids, to_ids)
        return self.bulk_mutation(op_builder_list)

    def bulk_unlink(self, from_ids, to_ids):
        """
        General method to bulk unlink GraphQL objects by id
        requests will be send in batch to avoid Payload Too Large exception
        :param from_ids: list of politylink ids
        :param to_ids: list of politylink ids
        :return: json data of the last request
        """

        op_builder_list = map(lambda x, y: partial(self.build_link_operation, from_id=x, to_id=y, remove=True),
                              from_ids, to_ids)
        return self.bulk_mutation(op_builder_list)

    def bulk_mutation(self, op_builder_list):
        ret = None
        op = Operation(Mutation)
        for op_builder in op_builder_list:
            op = op_builder(op=op)
            if len(op) >= self.batch_size:
                ret = self.exec(op)
                op = Operation(Mutation)
        if len(op):
            ret = self.exec(op)
        return ret

    def get_all_bills(self, fields=None, filter_=None, depth=1):
        """
        Special method to get all Bills
        :return: list of Bill
        """

        if fields is None:
            fields = ['id', 'name', 'bill_number']
        return self.get_all_objects('bill', fields, filter_, depth)

    def get_all_committees(self, fields=None, filter_=None, depth=1):
        """
        Special method to get all Committees
        :return: list of Committee
        """

        if fields is None:
            fields = ['id', 'name']
        return self.get_all_objects('committee', fields, filter_, depth)

    def get_all_minutes(self, fields=None, filter_=None, depth=1):
        """
        Special method to get all Minutes
        :return: list of Minutes
        """

        if fields is None:
            fields = ['id', 'name', 'start_date_time']
        return self.get_all_objects('minutes', fields, filter_, depth)

    def get_all_members(self, fields=None, filter_=None, depth=1):
        """
        Special method to get all Members
        :return: list of Member
        """

        if fields is None:
            fields = ['id', 'name']
        return self.get_all_objects('member', fields, filter_, depth)

    def get_all_diets(self, fields=None, filter_=None, depth=1):
        """
        Special method to get all Diets
        :return: list of Diet
        """

        if fields is None:
            fields = ['id', 'number']
        return self.get_all_objects('diet', fields, filter_, depth)

    def get_all_news(self, fields=None, filter_=None, depth=1, start_date=None, end_date=None):
        """
        Special method to get all News
        :return: list of News
        """

        def to_neo4j_datetime(dt):
            return _Neo4jDateTimeInput(year=dt.year, month=dt.month, day=dt.day)

        if fields is None:
            fields = ['id', 'title', 'published_at', 'url']
        if filter_ is None:
            filter_ = _NewsFilter(None)
        if start_date:
            filter_.published_at_gte = to_neo4j_datetime(start_date)
        if end_date:
            filter_.published_at_lt = to_neo4j_datetime(end_date)
        return self.get_all_objects('news', fields, filter_, depth)

    def get_all_activities(self, fields=None, filter_=None, depth=1, member_id=None, bill_id=None, minutes_id=None):
        """
        Special method to get all Activities
        :return: list of Activity
        """

        if fields is None:
            fields = ['id', 'datetime', 'member_id', 'bill_id', 'minutes_id']
        if filter_ is None:
            filter_ = _ActivityFilter(None)
        if member_id:
            filter_.member_id = member_id
        if bill_id:
            filter_.bill_id = bill_id
        if minutes_id:
            filter_.minutes_id = minutes_id
        return self.get_all_objects('activity', fields, filter_, depth)

    def get_all_objects(self, class_name, fields=None, filter_=None, depth=1):
        op = self.build_get_all_operation(class_name, fields, filter_)
        res = self.endpoint(op.__to_graphql__(auto_select_depth=depth))
        self.validate_response_or_raise(res)
        return getattr(op + res, class_name)

    @staticmethod
    def validate_response_or_raise(res):
        if 'errors' in res:
            raise GraphQLException(res['errors'])

    @staticmethod
    def build_merge_operation(obj, op=None):
        param = GraphQLClient.build_merge_param(obj)
        if op is None:
            op = Operation(Mutation)
        else:
            param['__alias__'] = f'op{len(op)}'

        method_name = 'merge_{}'.format(obj.id.split(':')[0].lower())
        try:
            res = getattr(op, method_name)(**param)
        except AttributeError:
            raise GraphQLException(f'unknown id type : {obj.id}')
        res.id()
        return op

    @staticmethod
    def build_merge_param(obj):
        param = dict()
        for field_name in obj.__field_names__:
            if hasattr(obj, field_name):
                param[field_name] = getattr(obj, field_name)
        return param

    @staticmethod
    def build_get_operation(id_, fields=None):
        op = Operation(Query)
        method_name = id_.split(':')[0].lower()
        try:
            obj = getattr(op, method_name)(filter=GraphQLClient.build_filter(id_))
        except AttributeError:
            raise GraphQLException(f'unknown id type : {id_}')
        if fields:
            for field in fields:
                getattr(obj, field)()
        return op

    @staticmethod
    def build_delete_operation(id_, op=None):
        if op is None:
            op = Operation(Mutation)
            maybe_alias = None
        else:
            maybe_alias = f'op{len(op)}'

        method_name = 'delete_{}'.format(id_.split(':')[0].lower())
        try:
            res = getattr(op, method_name)(id=id_, __alias__=maybe_alias)
        except AttributeError:
            raise GraphQLException(f'unknown id type : {id_}')
        res.id()
        return op

    @staticmethod
    def build_link_operation(from_id, to_id, remove=False, op=None):
        if op is None:
            op = Operation(Mutation)
            maybe_alias = None
        else:
            maybe_alias = f'op{len(op)}'

        method_name = GraphQLClient.build_link_method_name(from_id, to_id, remove)
        res = getattr(op, method_name)(
            from_=GraphQLClient.build_input(from_id),
            to=GraphQLClient.build_input(to_id),
            __alias__=maybe_alias
        )
        res.from_.id()
        res.to.id()
        return op

    @staticmethod
    def build_link_method_name(from_id, to_id, remove=False):
        from_id_type = from_id.split(':')[0]
        to_id_type = to_id.split(':')[0]
        key = (from_id_type, to_id_type)

        if key not in _link_method_name_map:
            raise GraphQLException(f'unknown id types to link: from={from_id} to={to_id}')

        method_prefix = 'remove' if remove else 'merge'
        method_body = _link_method_name_map[key]
        return f'{method_prefix}_{method_body}'

    @staticmethod
    def build_input(id_: str):
        # noinspection PyUnresolvedReferences
        from politylink.graphql.schema import _BillInput, _CommitteeInput, _MinutesInput, _SpeechInput, \
            _UrlInput, _NewsInput, _TimelineInput, _MemberInput, _DietInput, _ActivityInput

        class_name = '_{}Input'.format(id_.split(':')[0])
        try:
            return locals()[class_name]({'id': id_})
        except KeyError:
            raise GraphQLException(f'unknown id type to build GraphQL input: {id_}')

    @staticmethod
    def build_filter(id_: str):
        # noinspection PyUnresolvedReferences
        from politylink.graphql.schema import _BillFilter, _CommitteeFilter, _MinutesFilter, _SpeechFilter, \
            _UrlFilter, _NewsFilter, _TimelineFilter, _MemberFilter, _DietFilter, _ActivityFilter

        class_name = '_{}Filter'.format(id_.split(':')[0])
        try:
            return locals()[class_name]({'id': id_})
        except KeyError:
            raise GraphQLException(f'unknown id type to build GraphQL filter: {id_}')

    @staticmethod
    def build_get_all_operation(class_name, fields=None, filter_=None):
        op = Operation(Query)
        ret = getattr(op, class_name)(filter=filter_)
        if fields is not None:
            for field in fields:
                getattr(ret, field)()
        return op


# key1: from id type, key2: to id type, value: link method name
# we may automatically generate this map from schema.py
_link_method_name_map = {
    ('Url', 'Bill'): 'url_referred_bills',
    ('Url', 'Law'): 'url_referred_laws',
    ('Url', 'Member'): 'url_referred_members',
    ('Url', 'Minutes'): 'url_referred_minutes',
    ('Url', 'Activity'): 'url_referred_activities',
    ('News', 'Bill'): 'news_referred_bills',
    ('News', 'Law'): 'news_referred_laws',
    ('News', 'Member'): 'news_referred_members',
    ('News', 'Minutes'): 'news_referred_minutes',
    ('News', 'Timeline'): 'timeline_news',
    ('Activity', 'Member'): 'activity_member',
    ('Activity', 'Minutes'): 'activity_minutes',
    ('Activity', 'Bill'): 'activity_bill',
    ('Speech', 'Minutes'): 'speech_belonged_to_minutes',
    ('Minutes', 'Bill'): 'minutes_discussed_bills',
    ('Minutes', 'Law'): 'minutes_discussed_laws',
    ('Minutes', 'Committee'): 'minutes_belonged_to_committee',
    ('Minutes', 'Timeline'): 'timeline_minutes',
    ('Minutes', 'Diet'): 'minutes_belonged_to_diet',
    ('Bill', 'Committee'): 'bill_belonged_to_committees',
    ('Bill', 'Timeline'): 'timeline_bills',
    ('Bill', 'Diet'): 'bill_belonged_to_diets',
    ('Member', 'Bill'): 'member_submitted_bills',
    ('Member', 'Diet'): 'member_attended_diets',
    ('Member', 'Speech'): 'member_delivered_speeches',
    ('Member', 'Minutes'): 'member_attended_minutes'
}



import os
import re
from datetime import datetime

from kanjize import kanji2int


def get_env_or_default(key, default):
    return os.environ[key] if key in os.environ else default


class DateConverter:
    era2year = {'明治': 1872, '大正': 1912, '昭和': 1925, '平成': 1989, '令和': 2019}
    pattern = r'((?:明治|大正|昭和|平成|令和).*)年(.*)月(.*)日'

    @staticmethod
    def convert(s):
        """
        和暦の日付をdatetimeに変換する
        """

        m = re.search(DateConverter.pattern, s)
        if not m:
            raise ValueError(f'{s} does not fully match with {DateConverter.pattern}')
        year = DateConverter.convert_year_to_int(m.group(1))
        month = DateConverter.convert_to_int(m.group(2))
        day = DateConverter.convert_to_int(m.group(3))
        return datetime(year=year, month=month, day=day)

    @staticmethod
    def convert_year_to_int(s):
        s = s.strip()
        for era, year in DateConverter.era2year.items():
            if s.startswith(era):
                offset = DateConverter.convert_to_int(s[len(era):]) - 1
                return year + offset
        return DateConverter.convert_to_int(s)

    @staticmethod
    def convert_to_int(s):
        s = s.strip()
        if s.isdigit():
            return int(s)
        elif s == '元':
            return 1
        else:
            return kanji2int(s)



from politylink.graphql.client import GraphQLClient
from politylink.helpers.abstract_finder import AbstractFinder


class MemberFinder(AbstractFinder):
    """
    Cache based Member finder
    """

    def __init__(self, members=None, search_fields=None, **kwargs):
        super().__init__(search_fields or ['name', 'name_hira'])
        if members:
            self.members = members
        else:
            client = GraphQLClient(**kwargs)
            self.members = client.get_all_members(['id'] + self.search_fields)

    def find(self, text, exact_match=False, *args, **kwargs):
        return list(filter(lambda x: self.is_text_match(x, text, exact_match), self.members))



from politylink.graphql.client import GraphQLClient
from politylink.helpers.abstract_finder import AbstractFinder


class BillFinder(AbstractFinder):
    """
    Cache based Bill finder
    """

    def __init__(self, bills=None, search_fields=None, **kwargs):
        super().__init__(search_fields or ['name', 'bill_number', 'category'])
        if bills:
            self.bills = bills
        else:
            client = GraphQLClient(**kwargs)
            self.bills = client.get_all_bills(['id'] + self.search_fields)

    def find(self, text, exact_match=False, diet_number=None, category=None, *args, **kwargs):
        return list(filter(
            lambda x: self.is_text_match(x, text, exact_match) and
                      self.is_diet_match(x, diet_number) and
                      self.is_category_match(x, category),
            self.bills
        ))

    @staticmethod
    def is_diet_match(bill, diet_number=None):
        if diet_number is None:
            return True
        if hasattr(bill, 'bill_number'):
            return f'第{diet_number}回国会' in bill.bill_number
        return False

    @staticmethod
    def is_category_match(bill, category=None):
        if category is None:
            return True
        if hasattr(bill, 'category'):
            return bill.category == category
        return False



from urllib.parse import urlparse

import math
from elasticsearch import Elasticsearch

from politylink.elasticsearch.schema import AbstractText, to_cls


class ElasticsearchException(Exception):
    pass


class ElasticsearchClient:
    """
    Elasticsearch client for politylink endpoint
    """

    def __init__(self, url='http://localhost:9200'):
        def to_node(url):
            res = urlparse(url)
            return {'host': res.hostname, 'port': res.port}

        self.client = Elasticsearch(hosts=[to_node(url)])

    def index(self, obj):
        """
        create or update a document
        """

        assert isinstance(obj, AbstractText)
        try:
            return self.client.index(index=obj.index, id=obj.id, body=obj.__dict__)
        except Exception as e:
            raise ElasticsearchException(f'failed to index {obj}') from e

    def get(self, id_):
        """
        get a document by politylink id (ref idgen)
        """

        try:
            cls = to_cls(id_)
            res = self.client.get(index=cls.index, id=id_)
            return cls(res['_source'])
        except Exception as e:
            raise ElasticsearchException(f'failed to get {id_}') from e

    def search(self, cls, query=None):
        """
        search $cls documents by query
        return all documents when query is empty
        """

        if query:
            query_doc = {'query': {'multi_match': {'query': query, 'fields': cls.get_all_fields()}}}
        else:
            query_doc = {'query': {'match_all': {}}}
        try:
            res = self.client.search(index=cls.index, body=query_doc)
            return list(map(lambda hit: cls(hit['_source']), res['hits']['hits']))
        except Exception as e:
            raise ElasticsearchException(f'failed to search {cls.__class__.__name__} for {query_doc}') from e

    def get_term_statistics(self, id_):
        """
        get term statistics using termvectors API

        :param id_: politylink ID
        :return: dictionary of term to statistics (tf, idf, tfidf)
        """

        try:
            cls = to_cls(id_)
            res = self.client.termvectors(index=cls.index, id=id_, params={'fields': 'body', 'term_statistics': 'true'})
        except Exception as e:
            raise ElasticsearchException(f'failed to get termvectors for {id_}') from e

        term2stats = dict()
        doc_count = res['term_vectors']['body']['field_statistics']['doc_count']
        for term, stats_raw in res['term_vectors']['body']['terms'].items():
            stats = dict()
            stats['tf'] = stats_raw['term_freq']
            stats['idf'] = math.log(doc_count / stats_raw['doc_freq'])
            stats['tfidf'] = stats['tf'] * stats['idf']
            term2stats[term] = stats
        return term2stats



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['politylink',
 'politylink.elasticsearch',
 'politylink.graphql',
 'politylink.helpers',
 'politylink.nlp']

package_data = \
{'': ['*']}

install_requires = \
['elasticsearch>=7.9.1,<8.0.0',
 'ginza>=4.0.5,<5.0.0',
 'kanjize>=0.1.0,<0.2.0',
 'nltk>=3.5,<4.0',
 'pandas>=1.0.5,<2.0.0',
 'sgqlc>=11.0,<12.0',
 'spacy>=2.3.5,<3.0.0']

setup_kwargs = {
    'name': 'politylink',
    'version': '0.1.60',
    'description': '',
    'long_description': '## インストール\n```\npip install politylink\n```\n`politylink.nlp.keyphrase`を使用する場合は、追加で以下を実行してください。\n```\npip install git+https://github.com/boudinfl/pke.git\npython -m nltk.downloader stopwords\n```\n## 使い方\n\n### GraphQLClient\n\nPolityLinkの[GraphQLエンドポイント](https://graphql.politylink.jp/)にアクセスするためのGraphQLClientが用意されています。\n```\nfrom politylink.graphql.client import GraphQLClient\nclient = GraphQLClient()\n```\n\n#### 基本編\n\n`exec`メソッドを使えば任意のGraphQLクエリを実行することができます。\n```\nquery = """\nquery {\n  Bill(filter: {submittedDate: {year: 2020, month: 1, day: 20}}) {\n    name\n  }\n}\n"""\nclient.exec(query)\n```\n2020年1月20日に提出された3つの法律案の名前がJSON形式で得られるはずです。\n```\n{\'data\': {\'Bill\': [{\'name\': \'特定複合観光施設区域の整備の推進に関する法律及び特定複合観光施設区域整備法を廃止する法律案\'},\n   {\'name\': \'地方交付税法及び特別会計に関する法律の一部を改正する法律案\'},\n   {\'name\': \'平成三十年度歳入歳出の決算上の剰余金の処理の特例に関する法律案\'}]}}\n```\n\nまた、`get_all_*`メソッドを使うことで、返り値をJSONではなくPythonクラスのインスタンスとして取得することができます。\n例えば`get_all_bills`では法律案をBillインスタンスとして取得することができます。\n\n```\nbills = client.get_all_bills(fields=[\'id\', \'name\'])\nfirst_bill = bills[0]\n\nprint(f\'{len(bills)}件の法律案を取得しました\')\nprint(f\'最初の法律案は「{first_bill.name}」（{first_bill.id}）です\')\n```\n\n全ての法律案のidと名前（name）が得られるはずです。\n返り値はBillインスタンスなので、ドットを使って各フィールドにアクセスできます。\n```\n207件の法律案を取得しました\n最初の法律案は「地方交付税法及び特別会計に関する法律の一部を改正する法律案」（Bill:s1QZfjoCPyfdXXbrplP3-A）です\n```\n\nまた`filter_`を引数に渡すことで条件を指定して取得することもできます。詳しくは応用編を見てください。\n\n\n#### 応用編\n\nGraphQLClientは[sgqlc](https://github.com/profusion/sgqlc)のラッパークラスであり、クエリをコードで組み立てることも可能です。\n例えば最初の`exec`の例のクエリを組み立てると以下のようになります。\n\n```\nfrom politylink.graphql.schema import Query, _BillFilter, _Neo4jDateTimeInput\nfrom sgqlc.operation import Operation\n\nop = Operation(Query)\nfilter_ = _BillFilter(None)\nfilter_.submitted_date = _Neo4jDateTimeInput(year=2020, month=1, day=20)\nbills = op.bill(filter=filter_)\nbills.name()\nclient.exec(op)\n```\n\n組み立てた`Operation`は自動で文字列に変換されるので`exec`に直接渡すことができます。\n\nまた上で作ったfilter_は`get_all_*`メソッドの引数として渡すこともできます。\n```\nclient.get_all_bills(fields=[\'name\'], filter_=filter_)\n```\n\n最初の3件の法律案がBillインスタンスとして取得できました。\n```\n[Bill(name=\'地方交付税法及び特別会計に関する法律の一部を改正する法律案\'),\n Bill(name=\'平成三十年度歳入歳出の決算上の剰余金の処理の特例に関する法律案\'),\n Bill(name=\'特定複合観光施設区域の整備の推進に関する法律及び特定複合観光施設区域整備法を廃止する法律案\')]\n```\n',
    'author': 'Mitsuki Usui',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://politylink.jp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
